import asyncio

from aioredis.locks import Lock


async def test_finished_waiter_cancelled():
    lock = Lock()

    ta = asyncio.ensure_future(lock.acquire())
    await asyncio.sleep(0)
    assert lock.locked()

    tb = asyncio.ensure_future(lock.acquire())
    await asyncio.sleep(0)
    assert len(lock._waiters) == 1

    # Create a second waiter, wake up the first, and cancel it.
    # Without the fix, the second was not woken up and the lock
    # will never be locked
    asyncio.ensure_future(lock.acquire())
    await asyncio.sleep(0)
    lock.release()
    tb.cancel()

    await asyncio.sleep(0)
    assert ta.done()
    assert tb.cancelled()

    await asyncio.sleep(0)
    assert lock.locked()



from typing import Optional, Sequence   # noqa

__all__ = [
    'RedisError',
    'ProtocolError',
    'ReplyError',
    'MaxClientsError',
    'AuthError',
    'PipelineError',
    'MultiExecError',
    'WatchVariableError',
    'ChannelClosedError',
    'ConnectionClosedError',
    'ConnectionForcedCloseError',
    'PoolClosedError',
    'MasterNotFoundError',
    'SlaveNotFoundError',
    'ReadOnlyError',
    ]


class RedisError(Exception):
    """Base exception class for aioredis exceptions."""


class ProtocolError(RedisError):
    """Raised when protocol error occurs."""


class ReplyError(RedisError):
    """Raised for redis error replies (-ERR)."""

    MATCH_REPLY = None  # type: Optional[Sequence[str]]

    def __new__(cls, msg, *args):
        for klass in cls.__subclasses__():
            if msg and klass.MATCH_REPLY and msg.startswith(klass.MATCH_REPLY):
                return klass(msg, *args)
        return super().__new__(cls, msg, *args)


class MaxClientsError(ReplyError):
    """Raised for redis server when the maximum number of client has been
    reached."""

    MATCH_REPLY = "ERR max number of clients reached"


class AuthError(ReplyError):
    """Raised when authentication errors occurs."""

    MATCH_REPLY = (
        "NOAUTH ",
        "ERR invalid password",
        "ERR Client sent AUTH, but no password is set",
    )


class BusyGroupError(ReplyError):
    """Raised if Consumer Group name already exists."""

    MATCH_REPLY = "BUSYGROUP Consumer Group name already exists"


class PipelineError(RedisError):
    """Raised if command within pipeline raised error."""

    def __init__(self, errors):
        super().__init__('{} errors:'.format(self.__class__.__name__), errors)


class MultiExecError(PipelineError):
    """Raised if command within MULTI/EXEC block caused error."""


class WatchVariableError(MultiExecError):
    """Raised if watched variable changed (EXEC returns None)."""


class ChannelClosedError(RedisError):
    """Raised when Pub/Sub channel is unsubscribed and messages queue is empty.
    """


class ReadOnlyError(RedisError):
    """Raised from slave when read-only mode is enabled"""


class MasterNotFoundError(RedisError):
    """Raised for sentinel master not found error."""


class SlaveNotFoundError(RedisError):
    """Raised for sentinel slave not found error."""


class MasterReplyError(RedisError):
    """Raised by sentinel client for master error replies."""


class SlaveReplyError(RedisError):
    """Raised by sentinel client for slave error replies."""


class ConnectionClosedError(RedisError):
    """Raised if connection to server was closed."""


class ConnectionForcedCloseError(ConnectionClosedError):
    """Raised if connection was closed with .close() method."""


class PoolClosedError(RedisError):
    """Raised if pool is closed."""



import pytest

__all__ = [
    'redis_version',
]


def redis_version(*version, reason):
    assert 1 < len(version) <= 3, version
    assert all(isinstance(v, int) for v in version), version
    return pytest.mark.redis_version(version=version, reason=reason)



import pytest
import asyncio
import sys

from aioredis import (
    SlaveNotFoundError,
    ReadOnlyError,
    )
from _testutils import redis_version


pytestmark = redis_version(2, 8, 12, reason="Sentinel v2 required")
if sys.platform == 'win32':
    pytestmark = pytest.mark.skip(reason="unstable on windows")


@pytest.mark.timeout(40)
async def test_auto_failover(start_sentinel, start_server,
                             create_sentinel, create_connection):
    server1 = start_server('master-failover', ['slave-read-only yes'])
    start_server('slave-failover1', ['slave-read-only yes'], slaveof=server1)
    start_server('slave-failover2', ['slave-read-only yes'], slaveof=server1)

    sentinel1 = start_sentinel('sentinel-failover1', server1, quorum=2,
                               down_after_milliseconds=300,
                               failover_timeout=1000)
    sentinel2 = start_sentinel('sentinel-failover2', server1, quorum=2,
                               down_after_milliseconds=300,
                               failover_timeout=1000)
    # Wait a bit for sentinels to sync
    await asyncio.sleep(3)

    sp = await create_sentinel([sentinel1.tcp_address,
                                sentinel2.tcp_address],
                               timeout=1)

    _, old_port = await sp.master_address(server1.name)
    # ignoring host
    assert old_port == server1.tcp_address.port
    master = sp.master_for(server1.name)
    res = await master.role()
    assert res.role == 'master'
    assert master.address is not None
    assert master.address[1] == old_port

    # wait failover
    conn = await create_connection(server1.tcp_address)
    await conn.execute("debug", "sleep", 2)

    # _, new_port = await sp.master_address(server1.name)
    # assert new_port != old_port
    # assert new_port == server2.tcp_address.port
    assert (await master.set("key", "val"))
    assert master.address is not None
    assert master.address[1] != old_port


async def test_sentinel_normal(sentinel, create_sentinel):
    redis_sentinel = await create_sentinel([sentinel.tcp_address], timeout=1)
    redis = redis_sentinel.master_for('masterA')

    info = await redis.role()
    assert info.role == 'master'

    key, field, value = b'key:hset', b'bar', b'zap'
    exists = await redis.hexists(key, field)
    if exists:
        ret = await redis.hdel(key, field)
        assert ret != 1

    ret = await redis.hset(key, field, value)
    assert ret == 1
    ret = await redis.hset(key, field, value)
    assert ret == 0


@pytest.mark.xfail(reason="same sentinel; single master;")
async def test_sentinel_slave(sentinel, create_sentinel):
    redis_sentinel = await create_sentinel([sentinel.tcp_address], timeout=1)
    redis = redis_sentinel.slave_for('masterA')

    info = await redis.role()
    assert info.role == 'slave'

    key, field, value = b'key:hset', b'bar', b'zap'
    # redis = await get_slave_connection()
    exists = await redis.hexists(key, field)
    if exists:
        with pytest.raises(ReadOnlyError):
            await redis.hdel(key, field)

    with pytest.raises(ReadOnlyError):
        await redis.hset(key, field, value)


@pytest.mark.xfail(reason="Need proper sentinel configuration")
async def test_sentinel_slave_fail(sentinel, create_sentinel):
    redis_sentinel = await create_sentinel([sentinel.tcp_address], timeout=1)

    key, field, value = b'key:hset', b'bar', b'zap'

    redis = redis_sentinel.slave_for('masterA')
    exists = await redis.hexists(key, field)
    if exists:
        with pytest.raises(ReadOnlyError):
            await redis.hdel(key, field)

    with pytest.raises(ReadOnlyError):
        await redis.hset(key, field, value)

    ret = await redis_sentinel.failover('masterA')
    assert ret is True
    await asyncio.sleep(2)

    with pytest.raises(ReadOnlyError):
        await redis.hset(key, field, value)

    ret = await redis_sentinel.failover('masterA')
    assert ret is True
    await asyncio.sleep(2)
    while True:
        try:
            await asyncio.sleep(1)
            await redis.hset(key, field, value)
        except SlaveNotFoundError:
            continue
        except ReadOnlyError:
            break


@pytest.mark.xfail(reason="Need proper sentinel configuration")
async def test_sentinel_normal_fail(sentinel, create_sentinel):
    redis_sentinel = await create_sentinel([sentinel.tcp_address], timeout=1)

    key, field, value = b'key:hset', b'bar', b'zap'
    redis = redis_sentinel.master_for('masterA')
    exists = await redis.hexists(key, field)
    if exists:
        ret = await redis.hdel(key, field)
        assert ret == 1

    ret = await redis.hset(key, field, value)
    assert ret == 1
    ret = await redis_sentinel.failover('masterA')
    assert ret is True
    await asyncio.sleep(2)
    ret = await redis.hset(key, field, value)
    assert ret == 0
    ret = await redis_sentinel.failover('masterA')
    assert ret is True
    await asyncio.sleep(2)
    redis = redis_sentinel.slave_for('masterA')
    while True:
        try:
            await redis.hset(key, field, value)
            await asyncio.sleep(1)
            # redis = await get_slave_connection()
        except ReadOnlyError:
            break


@pytest.mark.timeout(30)
async def test_failover_command(start_server, start_sentinel,
                                create_sentinel):
    server = start_server('master-failover-cmd', ['slave-read-only yes'])
    start_server('slave-failover-cmd', ['slave-read-only yes'], slaveof=server)

    sentinel = start_sentinel('sentinel-failover-cmd', server, quorum=1,
                              down_after_milliseconds=300,
                              failover_timeout=1000)

    name = 'master-failover-cmd'
    redis_sentinel = await create_sentinel([sentinel.tcp_address], timeout=1)
    # Wait a bit for sentinels to sync
    await asyncio.sleep(3)

    orig_master = await redis_sentinel.master_address(name)
    assert await redis_sentinel.failover(name) is True
    await asyncio.sleep(2)

    new_master = await redis_sentinel.master_address(name)
    assert orig_master != new_master

    ret = await redis_sentinel.failover(name)
    assert ret is True
    await asyncio.sleep(2)

    new_master = await redis_sentinel.master_address(name)
    assert orig_master == new_master

    # This part takes almost 10 seconds (waiting for '+convert-to-slave').
    # Disabled for time being.

    # redis = redis_sentinel.slave_for(name)
    # while True:
    #     try:
    #         await asyncio.sleep(.2)
    #         await redis.set('foo', 'bar')
    #     except SlaveNotFoundError:
    #         pass
    #     except ReadOnlyError:
    #         break



import asyncio
import aioredis


async def main():
    sentinel = await aioredis.create_sentinel(
        ['redis://localhost:26379', 'redis://sentinel2:26379'])
    redis = sentinel.master_for('mymaster')

    ok = await redis.set('key', 'value')
    assert ok
    val = await redis.get('key', encoding='utf-8')
    assert val == 'value'

asyncio.run(main())




async def test_ssl_connection(create_connection, server, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    conn = await create_connection(
        ('localhost', ssl_port), ssl=ssl_ctx)
    res = await conn.execute('ping')
    assert res == b'PONG'


async def test_ssl_redis(create_redis, server, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    redis = await create_redis(
        ('localhost', ssl_port), ssl=ssl_ctx)
    res = await redis.ping()
    assert res == b'PONG'


async def test_ssl_pool(create_pool, server, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    pool = await create_pool(
        ('localhost', ssl_port), ssl=ssl_ctx)
    with (await pool) as conn:
        res = await conn.execute('PING')
        assert res == b'PONG'



import asyncio
import sys

from asyncio.locks import Lock as _Lock

# Fixes an issue with all Python versions that leaves pending waiters
# without being awakened when the first waiter is canceled.
# Code adapted from the PR https://github.com/python/cpython/pull/1031
# Waiting once it is merged to make a proper condition to relay on
# the stdlib implementation or this one patched


class Lock(_Lock):

    if sys.version_info < (3, 7, 0):
        async def acquire(self):
            """Acquire a lock.
            This method blocks until the lock is unlocked, then sets it to
            locked and returns True.
            """
            if not self._locked and all(w.cancelled() for w in self._waiters):
                self._locked = True
                return True

            fut = self._loop.create_future()

            self._waiters.append(fut)
            try:
                await fut
                self._locked = True
                return True
            except asyncio.CancelledError:
                if not self._locked:  # pragma: no cover
                    self._wake_up_first()
                raise
            finally:
                self._waiters.remove(fut)

        def _wake_up_first(self):
            """Wake up the first waiter who isn't cancelled."""
            for fut in self._waiters:
                if not fut.done():
                    fut.set_result(True)
                    break



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis(
        'redis://localhost')

    # No pipelining;
    async def wait_each_command():
        val = await redis.get('foo')    # wait until `val` is available
        cnt = await redis.incr('bar')   # wait until `cnt` is available
        return val, cnt

    # Sending multiple commands and then gathering results
    async def pipelined():
        fut1 = redis.get('foo')      # issue command and return future
        fut2 = redis.incr('bar')     # issue command and return future
        # block until results are available
        val, cnt = await asyncio.gather(fut1, fut2)
        return val, cnt

    # Explicit pipeline
    async def explicit_pipeline():
        pipe = redis.pipeline()
        fut1 = pipe.get('foo')
        fut2 = pipe.incr('bar')
        result = await pipe.execute()
        val, cnt = await asyncio.gather(fut1, fut2)
        assert result == [val, cnt]
        return val, cnt

    res = await wait_each_command()
    print(res)
    res = await pipelined()
    print(res)
    res = await explicit_pipeline()
    print(res)

    redis.close()
    await redis.wait_closed()


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')

    ch, = await redis.psubscribe('channel:*')
    assert isinstance(ch, aioredis.Channel)

    async def reader(channel):
        async for ch, message in channel.iter():
            print("Got message in channel:", ch, ":", message)
    asyncio.get_running_loop().create_task(reader(ch))

    await redis.publish('channel:1', 'Hello')
    await redis.publish('channel:2', 'World')

    redis.close()
    await redis.wait_closed()

asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')

    ch1, ch2 = await redis.subscribe('channel:1', 'channel:2')
    assert isinstance(ch1, aioredis.Channel)
    assert isinstance(ch2, aioredis.Channel)

    async def reader(channel):
        async for message in channel.iter():
            print("Got message:", message)
    asyncio.get_running_loop().create_task(reader(ch1))
    asyncio.get_running_loop().create_task(reader(ch2))

    await redis.publish('channel:1', 'Hello')
    await redis.publish('channel:2', 'World')

    redis.close()
    await redis.wait_closed()

asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')
    await redis.set('my-key', 'value')
    value = await redis.get('my-key', encoding='utf-8')
    print(value)

    redis.close()
    await redis.wait_closed()

asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')

    tr = redis.multi_exec()
    tr.set('key1', 'value1')
    tr.set('key2', 'value2')
    ok1, ok2 = await tr.execute()
    assert ok1
    assert ok2

asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')

    await redis.hmset_dict('hash',
                           key1='value1',
                           key2='value2',
                           key3=123)

    result = await redis.hgetall('hash', encoding='utf-8')
    assert result == {
        'key1': 'value1',
        'key2': 'value2',
        'key3': '123',  # note that Redis returns int as string
    }

    redis.close()
    await redis.wait_closed()

asyncio.run(main())



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis_pool('redis://localhost')
    await redis.set('key', 'string-value')
    bin_value = await redis.get('key')
    assert bin_value == b'string-value'

    str_value = await redis.get('key', encoding='utf-8')
    assert str_value == 'string-value'

    redis.close()
    await redis.wait_closed()

asyncio.run(main())
